import pytest
from yagrc import reflector as yagrc_reflector


@pytest.fixture(scope="module", params=["load_service", "load_all"])
def grpc_reflector(request, grpc_channel):
    reflector = yagrc_reflector.GrpcReflectionClient()
    if request.param == "load_service":
        reflector.load_protocols(grpc_channel, symbols=["Testing.Addition"])
    else:
        reflector.load_protocols(grpc_channel)
    return reflector


@pytest.fixture
def stub_class(grpc_reflector):
    return grpc_reflector.service_stub_class("Testing.Addition")


@pytest.fixture
def addend_class(grpc_reflector):
    return grpc_reflector.message_class("Testing.Addend")


@pytest.fixture
def query_class(grpc_reflector):
    return grpc_reflector.message_class("Testing.StatusQuery")


@pytest.fixture
def enum_class(grpc_reflector):
    return grpc_reflector.enum_class("Testing.ServiceState")


def test_AddOne(grpc_channel, stub_class, addend_class):
    stub = stub_class(grpc_channel)
    response = stub.AddOne(addend_class(number=5))
    assert response.number == 6


def test_AddOnes(grpc_channel, stub_class, addend_class):
    stub = stub_class(grpc_channel)
    responses = stub.AddOnes(addend_class(number=5))
    for i, response in enumerate(responses):
        assert response.number == 6 + i
    assert i == 4


def test_AddsOne(grpc_channel, stub_class, addend_class):
    stub = stub_class(grpc_channel)
    response = stub.AddsOne((addend_class(number=x) for x in range(5)))
    assert response.number == 0 + 1 + 2 + 3 + 4 + 1


def test_AddsOnes(grpc_channel, stub_class, addend_class):
    stub = stub_class(grpc_channel)
    responses = stub.AddsOnes((addend_class(number=x) for x in range(5)))
    for i, response in enumerate(responses):
        assert response.number == i + 1
    assert i == 4


def test_CountQuery(grpc_channel, stub_class, query_class):
    stub = stub_class(grpc_channel)
    response = stub.Status(query_class(count={}))
    assert response.count.number > 0


def test_StatusQuery(grpc_channel, stub_class, query_class, enum_class):
    stub = stub_class(grpc_channel)
    response = stub.Status(query_class(state={}))
    assert response.state.states["sum_engine"].state == enum_class.READY
    assert enum_class.Name(response.state.states["difference_engine"].state
                          ) == "BUSY_COMPUTING_ONE"
    # no reflector equivalent for AddTypes_pb2._UNKNOWN


def test_nested_messsage(grpc_reflector):
    nested_class = grpc_reflector.message_class(
        "Testing.StatusResponse.State.Wrapper")
    assert nested_class.DESCRIPTOR.full_name == "Testing.StatusResponse.State.Wrapper"


def test_list_services(grpc_channel):
    services = yagrc_reflector.list_services(grpc_channel)
    assert "Testing.Addition" in services


@pytest.mark.xfail(raises=yagrc_reflector.ServiceError, strict=True)
def test_list_error(grpc_channel, force_list_services_error):
    reflector = yagrc_reflector.GrpcReflectionClient()
    reflector.load_protocols(grpc_channel)


@pytest.mark.xfail(raises=KeyError, strict=True)
def test_unsatisfied_dep(grpc_channel, force_unsatisfied_dep):
    reflector = yagrc_reflector.GrpcReflectionClient()
    reflector.load_protocols(grpc_channel, filenames=["Testing_protos/Add_One.proto"])
    assert False


@pytest.mark.xfail(raises=yagrc_reflector.ServiceError, strict=True)
def test_bad_proto(grpc_channel):
    reflector = yagrc_reflector.GrpcReflectionClient()
    reflector.load_protocols(grpc_channel, symbols=["Testing.Subtraction"])
    assert False


@pytest.mark.xfail(raises=KeyError, strict=True)
def test_bad_service(grpc_reflector):
    grpc_reflector.service_stub_class("Testing.Subtraction")
    assert False


@pytest.mark.xfail(raises=KeyError, strict=True)
def test_bad_message(grpc_reflector):
    grpc_reflector.message_class("Testing.Minuend")
    assert False



import os.path
import sys

from grpc_reflection.v1alpha import reflection
from grpc_reflection.v1alpha import reflection_pb2
import pytest

# Testing_protos package is deliberately kept out of the import path, but the
# service still needs it. Can't just use relative imports, because the recursive
# imports in the generated _pb2 modules also need to be able to find it.
addone_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                           "service")
sys.path.insert(0, addone_path)

# See comments in Add_One.proto about terrible naming
from Testing_protos import Add_One_pb2_grpc as _service_Add_One_pb2_grpc
from Testing_protos.Add_One_pb2 import DESCRIPTOR as _Add_One_DESCRIPTOR
import addone_server

sys.path.remove(addone_path)
for name in ("Testing_protos", "Testing_protos.AddAltTypes_pb2",
             "Testing_protos.Add_One_pb2_grpc", "Testing_protos.AddTypes_pb2",
             "Testing_protos.Add_One_pb2"):
    if name in sys.modules:
        del sys.modules[name]


@pytest.fixture(scope='module')
def grpc_add_to_server():
    return _service_Add_One_pb2_grpc.add_AdditionServicer_to_server


@pytest.fixture(scope='module')
def grpc_servicer():
    return addone_server.Addition()


# grpc_stub_cls fixture not defined because stub creation is part of test


# Override pytest-grpc plugin implementation to enable reflection
@pytest.fixture(scope='module')
def grpc_server(_grpc_server, grpc_addr, grpc_add_to_server, grpc_servicer):
    grpc_add_to_server(grpc_servicer, _grpc_server)
    service_names = (
        _Add_One_DESCRIPTOR.services_by_name['Addition'].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(service_names, _grpc_server)
    _grpc_server.add_insecure_port(grpc_addr)
    _grpc_server.start()
    yield _grpc_server
    _grpc_server.stop(grace=None)


# Override pytest-grpc plugin implementation of grpc_channel to hack around
# unhandled optional args to handler methods. This is ugly as sin, only
# patches the one handler type the reflection service uses, and is liable to
# break with later versions of the plugin, but whatever.... On the plus side,
# this provides an opportunity to verify that reflection service operations
# always include a timeout arg.
@pytest.fixture(scope='module')
def grpc_channel(grpc_create_channel):
    with grpc_create_channel() as channel:
        real_fake_stream_stream = channel.stream_stream

        def wrapped_stream_stream(method_path, *args, **kwargs):
            real_fake_handler = real_fake_stream_stream(method_path, *args,
                                                        **kwargs)

            def wrapped_fake_handler(request, timeout=None, **kwargs):
                if method_path.startswith(
                        "/grpc.reflection.v1alpha.ServerReflection/"):
                    assert timeout is not None
                return real_fake_handler(request)

            return wrapped_fake_handler

        if channel.__class__.__name__ == "FakeChannel":
            channel.stream_stream = wrapped_stream_stream

        yield channel


# Brittle hack of the reflection service to report error
@pytest.fixture
def force_list_services_error(monkeypatch):

    def mock_error(self):
        response = reflection_pb2.ServerReflectionResponse()
        response.error_response.error_code = 1
        response.error_response.error_message = "fake error"
        return response

    monkeypatch.setattr(reflection.BaseReflectionServicer, "_list_services",
                        mock_error)


# And another to simulate an unsatisfied dependency
@pytest.fixture
def force_unsatisfied_dep(monkeypatch):
    real_method = reflection.BaseReflectionServicer._file_by_filename

    def mock_file_by_filename(self, filename):
        response = real_method(self, filename)
        del response.file_descriptor_response.file_descriptor_proto[-1]
        return response

    monkeypatch.setattr(reflection.BaseReflectionServicer, "_file_by_filename",
                        mock_file_by_filename)



import importlib
import sys

import pytest
from yagrc import importer as yagrc_importer

# See comments in Add_One.proto about terrible naming


# lazy import intentionally has no cleanup mechanism, so do it the hard way
@pytest.fixture
def lazy_cleaner():
    yield
    if yagrc_importer._lazy_finder in sys.meta_path:
        sys.meta_path.remove(yagrc_importer._lazy_finder)
    yagrc_importer._lazy_finder._package_modules.clear()
    yagrc_importer._lazy_finder.pb2_imports.clear()
    yagrc_importer._lazy_finder.pb2_grpc_imports.clear()
    yagrc_importer._lazy_importer.deconfigure()
    for name in ("Testing_protos", "Testing_protos.other_pb2",
                 "Testing_protos.Add_One_pb2_grpc",
                 "Testing_protos.AddTypes_pb2"):
        if name in sys.modules:
            del sys.modules[name]


def test_AddOne(grpc_channel, lazy_cleaner):
    yagrc_importer.add_lazy_packages(["Testing_protos"])
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    yagrc_importer.resolve_lazy_imports(grpc_channel)
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.AddOne(AddTypes_pb2.Addend(number=5))
    assert response.number == 6


def test_unresolved(lazy_cleaner):
    yagrc_importer.add_lazy_packages(["Testing_protos"])
    from Testing_protos import other_pb2
    try:
        a = other_pb2.random_attribute
        assert False
    except AttributeError as e:
        assert str(e).find("unresolved lazy module") >= 0


def test_unlazy(grpc_channel, lazy_cleaner):
    yagrc_importer.add_lazy_packages(["Testing_protos"])
    import collections
    yagrc_importer.resolve_lazy_imports(grpc_channel)


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_bad_import(lazy_cleaner):
    yagrc_importer.add_lazy_packages(["Testing_protos"])
    from Testing_protos import bad_import
    assert False


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_bad_package(lazy_cleaner):
    yagrc_importer.add_lazy_packages(["Testing_protos"])
    from bad_package import bad_import
    assert False


def test_resolve_fail(grpc_channel, lazy_cleaner):
    yagrc_importer.add_lazy_packages(["Testing_protos"])
    from Testing_protos import AddTypes_pb2

    try:
        yagrc_importer.resolve_lazy_imports(None)
        assert False
    except AttributeError:
        pass

    # Test failure again to verify first failure did not lose state.
    try:
        yagrc_importer.resolve_lazy_imports(None)
        assert False
    except AttributeError:
        pass

    # Finally, test success case to make sure prior failures did not break
    # things permanently, including the ability to do more lazy imports.
    from Testing_protos import Add_One_pb2_grpc
    yagrc_importer.resolve_lazy_imports(grpc_channel)
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.AddOne(AddTypes_pb2.Addend(number=5))
    assert response.number == 6



#!/usr/bin/python3
"""Non-reflection client for the addone_server test service.

This is not meant to be useful. See the clients in the examples directory for
slightly more useful grpc client code.
"""

import grpc

# See comments in Add_One.proto about terrible naming
from Testing_protos import Add_One_pb2_grpc
from Testing_protos import AddTypes_pb2


def main():
    with grpc.insecure_channel("localhost:19001") as channel:
        stub = Add_One_pb2_grpc.AdditionStub(channel)

        response = stub.AddOne(AddTypes_pb2.Addend(number=5))
        print(response)

        responses = stub.AddOnes(AddTypes_pb2.Addend(number=5))
        print(*responses)

        response = stub.AddsOne(
            (AddTypes_pb2.Addend(number=x) for x in range(5)))
        print(response)

        responses = stub.AddsOnes(
            (AddTypes_pb2.Addend(number=x) for x in range(5)))
        print(*responses)

        response = stub.Status(AddTypes_pb2.StatusQuery(count={}))
        print(response)

        response = stub.Status(AddTypes_pb2.StatusQuery(state={}))
        print(response)


if __name__ == '__main__':
    logging.basicConfig()
    main()



#!/usr/bin/python3
"""`yagrc.reflector` client for the subtract_server example service."""
import logging

import grpc
from yagrc import reflector as yagrc_reflector


def main():
    reflector = yagrc_reflector.GrpcReflectionClient()

    with grpc.insecure_channel("localhost:19002") as channel:
        reflector.load_protocols(channel, symbols=["Arithmetic.Subtraction"])
        stub_class = reflector.service_stub_class("Arithmetic.Subtraction")
        request_class = reflector.message_class("Arithmetic.Minuend")

        stub = stub_class(channel)
        response = stub.SubtractOne(request_class(number=5))
    print(response)


if __name__ == '__main__':
    logging.basicConfig()
    main()



#!/usr/bin/python3
"""Server for the ridiculous Addition service.

This is meant only for running this project's test cases against.
"""
from concurrent import futures
import logging

import grpc
from grpc_reflection.v1alpha import reflection

# See comments in Add_One.proto about terrible naming
from Testing_protos import Add_One_pb2
from Testing_protos import Add_One_pb2_grpc
from Testing_protos import AddTypes_pb2


class Addition(Add_One_pb2_grpc.AdditionServicer):

    def __init__(self):
        super().__init__()
        self.count = 0

    def AddOne(self, request, context):
        self.count += 1
        return AddTypes_pb2.Sum(number=request.number + 1)

    def AddOnes(self, request, context):
        self.count += 1
        return (AddTypes_pb2.Sum(number=request.number + x + 1)
                for x in range(request.number))

    def AddsOne(self, request_iterator, context):
        self.count += 1
        return AddTypes_pb2.Sum(
            number=sum(request.number for request in request_iterator) + 1)

    def AddsOnes(self, request_iterator, context):
        self.count += 1
        return (AddTypes_pb2.Sum(number=request.number + 1)
                for request in request_iterator)

    def Status(self, request, context):
        self.count += 1
        if request.HasField("count"):
            return AddTypes_pb2.StatusResponse(count={"number": self.count})
        if request.HasField("state"):
            response = AddTypes_pb2.StatusResponse()
            response.state.states[
                "sum_engine"].state = AddTypes_pb2.ServiceState.READY
            response.state.states[
                "difference_engine"].state = AddTypes_pb2.ServiceState.BUSY_COMPUTING_ONE
            response.state.states[
                "chaos_engine"].state = AddTypes_pb2.ServiceState._UNKNOWN
            return response
        raise NotImplementedError("request not implemented: " +
                                  request.WhichOneof("query"))


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    Add_One_pb2_grpc.add_AdditionServicer_to_server(Addition(), server)
    service_names = (
        Add_One_pb2.DESCRIPTOR.services_by_name['Addition'].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(service_names, server)
    server.add_insecure_port('[::]:19001')
    server.start()
    server.wait_for_termination()


if __name__ == '__main__':
    logging.basicConfig()
    serve()



#!/usr/bin/python3
"""`yagrc.importer` client for the subtract_server example service."""
import logging

import grpc
from yagrc import importer as yagrc_importer


def main():
    importer = yagrc_importer.GrpcImporter()

    with grpc.insecure_channel("localhost:19002") as channel:
        importer.configure(channel, filenames=["arithmetic/subtract.proto"])

        from arithmetic import subtract_pb2
        from arithmetic import subtract_pb2_grpc

        stub = subtract_pb2_grpc.SubtractionStub(channel)
        response = stub.SubtractOne(subtract_pb2.Minuend(number=5))
    print(response)


if __name__ == '__main__':
    logging.basicConfig()
    main()



#!/usr/bin/python3
"""Lazy `yagrc.importer` client for the subtract_server example service."""
import logging

import grpc
from yagrc import importer as yagrc_importer

yagrc_importer.add_lazy_packages(["arithmetic"])

from arithmetic import subtract_pb2
from arithmetic import subtract_pb2_grpc


def main():
    with grpc.insecure_channel("localhost:19002") as channel:
        yagrc_importer.resolve_lazy_imports(channel)

        stub = subtract_pb2_grpc.SubtractionStub(channel)
        response = stub.SubtractOne(subtract_pb2.Minuend(number=5))
    print(response)


if __name__ == '__main__':
    logging.basicConfig()
    main()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from arithmetic import subtract_pb2 as arithmetic_dot_subtract__pb2


class SubtractionStub(object):
    """A service for subtracting integer numbers
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SubtractOne = channel.unary_unary(
                '/Arithmetic.Subtraction/SubtractOne',
                request_serializer=arithmetic_dot_subtract__pb2.Minuend.SerializeToString,
                response_deserializer=arithmetic_dot_subtract__pb2.Difference.FromString,
                )


class SubtractionServicer(object):
    """A service for subtracting integer numbers
    """

    def SubtractOne(self, request, context):
        """Subtract 1 from a number
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SubtractionServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SubtractOne': grpc.unary_unary_rpc_method_handler(
                    servicer.SubtractOne,
                    request_deserializer=arithmetic_dot_subtract__pb2.Minuend.FromString,
                    response_serializer=arithmetic_dot_subtract__pb2.Difference.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Arithmetic.Subtraction', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Subtraction(object):
    """A service for subtracting integer numbers
    """

    @staticmethod
    def SubtractOne(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Arithmetic.Subtraction/SubtractOne',
            arithmetic_dot_subtract__pb2.Minuend.SerializeToString,
            arithmetic_dot_subtract__pb2.Difference.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: arithmetic/subtract.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x61rithmetic/subtract.proto\x12\nArithmetic\"\x19\n\x07Minuend\x12\x0e\n\x06number\x18\x01 \x01(\x05\"\x1c\n\nDifference\x12\x0e\n\x06number\x18\x01 \x01(\x05\x32I\n\x0bSubtraction\x12:\n\x0bSubtractOne\x12\x13.Arithmetic.Minuend\x1a\x16.Arithmetic.Differenceb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'arithmetic.subtract_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_MINUEND']._serialized_start=41
  _globals['_MINUEND']._serialized_end=66
  _globals['_DIFFERENCE']._serialized_start=68
  _globals['_DIFFERENCE']._serialized_end=96
  _globals['_SUBTRACTION']._serialized_start=98
  _globals['_SUBTRACTION']._serialized_end=171
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from Testing_protos import AddAltTypes_pb2 as Testing__protos_dot_AddAltTypes__pb2
from Testing_protos import AddTypes_pb2 as Testing__protos_dot_AddTypes__pb2


class AdditionStub(object):
    """A service for adding integer numbers
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddOne = channel.unary_unary(
                '/Testing.Addition/AddOne',
                request_serializer=Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
                response_deserializer=Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
                )
        self.AddOnes = channel.unary_stream(
                '/Testing.Addition/AddOnes',
                request_serializer=Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
                response_deserializer=Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
                )
        self.AddsOne = channel.stream_unary(
                '/Testing.Addition/AddsOne',
                request_serializer=Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
                response_deserializer=Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
                )
        self.AddsOnes = channel.stream_stream(
                '/Testing.Addition/AddsOnes',
                request_serializer=Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
                response_deserializer=Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
                )
        self.Status = channel.unary_unary(
                '/Testing.Addition/Status',
                request_serializer=Testing__protos_dot_AddTypes__pb2.StatusQuery.SerializeToString,
                response_deserializer=Testing__protos_dot_AddTypes__pb2.StatusResponse.FromString,
                )


class AdditionServicer(object):
    """A service for adding integer numbers
    """

    def AddOne(self, request, context):
        """Add 1 to a number
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddOnes(self, request, context):
        """Add 1 to a number multiple times
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddsOne(self, request_iterator, context):
        """Add 1 to a sum of numbers
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddsOnes(self, request_iterator, context):
        """Add 1 to multiple numbers
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Status(self, request, context):
        """Introspect the service
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AdditionServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddOne': grpc.unary_unary_rpc_method_handler(
                    servicer.AddOne,
                    request_deserializer=Testing__protos_dot_AddAltTypes__pb2.Addend.FromString,
                    response_serializer=Testing__protos_dot_AddAltTypes__pb2.Sum.SerializeToString,
            ),
            'AddOnes': grpc.unary_stream_rpc_method_handler(
                    servicer.AddOnes,
                    request_deserializer=Testing__protos_dot_AddAltTypes__pb2.Addend.FromString,
                    response_serializer=Testing__protos_dot_AddAltTypes__pb2.Sum.SerializeToString,
            ),
            'AddsOne': grpc.stream_unary_rpc_method_handler(
                    servicer.AddsOne,
                    request_deserializer=Testing__protos_dot_AddAltTypes__pb2.Addend.FromString,
                    response_serializer=Testing__protos_dot_AddAltTypes__pb2.Sum.SerializeToString,
            ),
            'AddsOnes': grpc.stream_stream_rpc_method_handler(
                    servicer.AddsOnes,
                    request_deserializer=Testing__protos_dot_AddAltTypes__pb2.Addend.FromString,
                    response_serializer=Testing__protos_dot_AddAltTypes__pb2.Sum.SerializeToString,
            ),
            'Status': grpc.unary_unary_rpc_method_handler(
                    servicer.Status,
                    request_deserializer=Testing__protos_dot_AddTypes__pb2.StatusQuery.FromString,
                    response_serializer=Testing__protos_dot_AddTypes__pb2.StatusResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Testing.Addition', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Addition(object):
    """A service for adding integer numbers
    """

    @staticmethod
    def AddOne(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Testing.Addition/AddOne',
            Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
            Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddOnes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/Testing.Addition/AddOnes',
            Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
            Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddsOne(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/Testing.Addition/AddsOne',
            Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
            Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddsOnes(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/Testing.Addition/AddsOnes',
            Testing__protos_dot_AddAltTypes__pb2.Addend.SerializeToString,
            Testing__protos_dot_AddAltTypes__pb2.Sum.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Status(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Testing.Addition/Status',
            Testing__protos_dot_AddTypes__pb2.StatusQuery.SerializeToString,
            Testing__protos_dot_AddTypes__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Testing_protos/AddAltTypes.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n Testing_protos/AddAltTypes.proto\x12\x07Testing\"\x18\n\x06\x41\x64\x64\x65nd\x12\x0e\n\x06number\x18\x01 \x01(\x05\"\x15\n\x03Sum\x12\x0e\n\x06number\x18\x01 \x01(\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'Testing_protos.AddAltTypes_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_ADDEND']._serialized_start=45
  _globals['_ADDEND']._serialized_end=69
  _globals['_SUM']._serialized_start=71
  _globals['_SUM']._serialized_end=92
# @@protoc_insertion_point(module_scope)



"""Utility function to dump FileDescriptorSet protocol buffers."""
from google.protobuf import descriptor_pb2

from yagrc import reflector


def dump_protocols(channel, filenames=None, symbols=None):
    """Get serialized FileDescriptorSet message via reflection.

    See `yagrc.reflector.GrpcReflectionClient.load_protocols` for arg and
    exception details.
    """
    fdset = descriptor_pb2.FileDescriptorSet()
    grpc_reflector = reflector.GrpcReflectionEngine()
    for filename in grpc_reflector.load_protocols(channel, filenames, symbols):
        proto = descriptor_pb2.FileDescriptorProto()
        descr = grpc_reflector.file_descriptor(filename)
        descr.CopyToProto(proto)
        fdset.file.append(proto)  # pylint: disable=no-member
    return fdset.SerializeToString()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Testing_protos/Add_One.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from Testing_protos import AddTypes_pb2 as Testing__protos_dot_AddTypes__pb2
try:
  Testing__protos_dot_AddAltTypes__pb2 = Testing__protos_dot_AddTypes__pb2.Testing__protos_dot_AddAltTypes__pb2
except AttributeError:
  Testing__protos_dot_AddAltTypes__pb2 = Testing__protos_dot_AddTypes__pb2.Testing_protos.AddAltTypes_pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cTesting_protos/Add_One.proto\x12\x07Testing\x1a\x1dTesting_protos/AddTypes.proto2\xf3\x01\n\x08\x41\x64\x64ition\x12\'\n\x06\x41\x64\x64One\x12\x0f.Testing.Addend\x1a\x0c.Testing.Sum\x12*\n\x07\x41\x64\x64Ones\x12\x0f.Testing.Addend\x1a\x0c.Testing.Sum0\x01\x12*\n\x07\x41\x64\x64sOne\x12\x0f.Testing.Addend\x1a\x0c.Testing.Sum(\x01\x12-\n\x08\x41\x64\x64sOnes\x12\x0f.Testing.Addend\x1a\x0c.Testing.Sum(\x01\x30\x01\x12\x37\n\x06Status\x12\x14.Testing.StatusQuery\x1a\x17.Testing.StatusResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'Testing_protos.Add_One_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_ADDITION']._serialized_start=73
  _globals['_ADDITION']._serialized_end=316
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Testing_protos/AddTypes.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from Testing_protos import AddAltTypes_pb2 as Testing__protos_dot_AddAltTypes__pb2

from Testing_protos.AddAltTypes_pb2 import *

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dTesting_protos/AddTypes.proto\x12\x07Testing\x1a Testing_protos/AddAltTypes.proto\"\x84\x01\n\x0bStatusQuery\x12,\n\x05\x63ount\x18\xe9\x07 \x01(\x0b\x32\x1a.Testing.StatusQuery.CountH\x00\x12,\n\x05state\x18\xea\x07 \x01(\x0b\x32\x1a.Testing.StatusQuery.StateH\x00\x1a\x07\n\x05\x43ount\x1a\x07\n\x05StateB\x07\n\x05query\"\xe3\x02\n\x0eStatusResponse\x12/\n\x05\x63ount\x18\xd1\x0f \x01(\x0b\x32\x1d.Testing.StatusResponse.CountH\x00\x12/\n\x05state\x18\xd2\x0f \x01(\x0b\x32\x1d.Testing.StatusResponse.StateH\x00\x1a\x17\n\x05\x43ount\x12\x0e\n\x06number\x18\x01 \x01(\r\x1a\xc9\x01\n\x05State\x12\x39\n\x06states\x18\x01 \x03(\x0b\x32).Testing.StatusResponse.State.StatesEntry\x1a/\n\x07Wrapper\x12$\n\x05state\x18\x01 \x01(\x0e\x32\x15.Testing.ServiceState\x1aT\n\x0bStatesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32%.Testing.StatusResponse.State.Wrapper:\x02\x38\x01\x42\n\n\x08response*?\n\x0cServiceState\x12\x0c\n\x08_UNKNOWN\x10\x00\x12\x16\n\x12\x42USY_COMPUTING_ONE\x10\x01\x12\t\n\x05READY\x10\x02P\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'Testing_protos.AddTypes_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATUSRESPONSE_STATE_STATESENTRY._options = None
  _STATUSRESPONSE_STATE_STATESENTRY._serialized_options = b'8\001'
  _globals['_SERVICESTATE']._serialized_start=569
  _globals['_SERVICESTATE']._serialized_end=632
  _globals['_STATUSQUERY']._serialized_start=77
  _globals['_STATUSQUERY']._serialized_end=209
  _globals['_STATUSQUERY_COUNT']._serialized_start=184
  _globals['_STATUSQUERY_COUNT']._serialized_end=191
  _globals['_STATUSQUERY_STATE']._serialized_start=193
  _globals['_STATUSQUERY_STATE']._serialized_end=200
  _globals['_STATUSRESPONSE']._serialized_start=212
  _globals['_STATUSRESPONSE']._serialized_end=567
  _globals['_STATUSRESPONSE_COUNT']._serialized_start=328
  _globals['_STATUSRESPONSE_COUNT']._serialized_end=351
  _globals['_STATUSRESPONSE_STATE']._serialized_start=354
  _globals['_STATUSRESPONSE_STATE']._serialized_end=555
  _globals['_STATUSRESPONSE_STATE_WRAPPER']._serialized_start=422
  _globals['_STATUSRESPONSE_STATE_WRAPPER']._serialized_end=469
  _globals['_STATUSRESPONSE_STATE_STATESENTRY']._serialized_start=471
  _globals['_STATUSRESPONSE_STATE_STATESENTRY']._serialized_end=555
# @@protoc_insertion_point(module_scope)



#!/usr/bin/python3
"""Non-reflection client for the subtract_server example service.

For comparison with the clients using `yagrc` for reflection.
"""
import logging

import grpc

from arithmetic import subtract_pb2
from arithmetic import subtract_pb2_grpc


def main():
    with grpc.insecure_channel("localhost:19002") as channel:
        stub = subtract_pb2_grpc.SubtractionStub(channel)
        response = stub.SubtractOne(subtract_pb2.Minuend(number=5))
    print(response)


if __name__ == '__main__':
    logging.basicConfig()
    main()



"""A gRPC reflection client.

This module provides a programmatic interface for getting grpc class types
discovered via gRPC reflection service.

See:
    https://grpc.io/ for more information on gRPC in general.

    https://github.com/grpc/grpc/blob/master/doc/server-reflection.md for more
    information on gRPC service reflection.
"""
from collections import deque

from google.protobuf import descriptor_pb2
from google.protobuf import descriptor_pool
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import message_factory
from grpc_reflection.v1alpha import reflection_pb2
from grpc_reflection.v1alpha import reflection_pb2_grpc

# This is somewhat arbitrary, and is just here to prevent hang in the case of
# dead network connection that the client side still believes to be open.
QUERY_TIMEOUT = 10
"""Max wait time for reflection query completion, in seconds."""


class ServiceError(Exception):
    """Error reported by reflection service."""


def __stub_init__(self, channel):
    """The __init__ method for service stub classes."""
    for name, call, full_name, out_serializer, in_serializer in self._methods:  # pylint: disable=protected-access
        setattr(
            self, name,
            getattr(channel, call)(full_name,
                                   request_serializer=out_serializer,
                                   response_deserializer=in_serializer))


def _list_services(stub):
    responses = stub.ServerReflectionInfo(iter(
        [reflection_pb2.ServerReflectionRequest(list_services="")]),
                                          timeout=QUERY_TIMEOUT)
    for response in responses:
        if response.HasField("error_response"):
            raise ServiceError(response.error_response.error_message)
        for service in response.list_services_response.service:
            yield service.name


def list_services(channel):
    """Get list of fully qualified service names via reflection.

    Args:
        channel (grpc.Channel): The RPC channel to use.

    Raises:
        ServiceError: Reflection service responded with an error.
        grpc.RpcError: Lower level RPC error.
    """
    stub = reflection_pb2_grpc.ServerReflectionStub(channel)
    return list(_list_services(stub))


def enum_from_descr(proto):
    """Get enum class from enum descriptor.

    Args:
        proto (google.protobuf.descriptor.EnumDescriptor): Enum descriptor.
        """
    return enum_type_wrapper.EnumTypeWrapper(proto)


class GrpcReflectionEngine():
    """Implementation class for gRPC reflection client.

    This class is not meant to be used directly. Instead, use
    `GrpcReflectionClient` or `yagrc.importer.GrpcImporter`
    """

    def __init__(self):
        self.methods_by_file = {}
        self.pool = descriptor_pool.DescriptorPool()

    def load_protocols(self, channel, filenames=None, symbols=None):
        """Implementation of `GrpcReflectionClient.load_protocols`"""
        stub = reflection_pb2_grpc.ServerReflectionStub(channel)

        requests = []
        if filenames:
            requests.extend(
                reflection_pb2.ServerReflectionRequest(file_by_filename=name)
                for name in filenames)
        if symbols:
            requests.extend(
                reflection_pb2.ServerReflectionRequest(
                    file_containing_symbol=symbol) for symbol in symbols)
        if not requests:
            requests.extend(
                reflection_pb2.ServerReflectionRequest(
                    file_containing_symbol=name)
                for name in _list_services(stub)
                if name != "grpc.reflection.v1alpha.ServerReflection")

        protos = {}
        traversed = set()
        while requests:
            responses = stub.ServerReflectionInfo(iter(requests),
                                                  timeout=QUERY_TIMEOUT)
            deps = set()
            for response in responses:
                if response.HasField("error_response"):
                    raise ServiceError(response.error_response.error_message)
                for desc_bytes in response.file_descriptor_response.file_descriptor_proto:
                    proto = descriptor_pb2.FileDescriptorProto.FromString(  # pylint: disable=no-member
                        desc_bytes)
                    traversed.add(proto.name)
                    deps.update(proto.dependency)
                    protos[proto.name] = proto
                    self.methods_by_file[proto.name] = {
                        service.name: service.method
                        for service in proto.service
                    }
            deps -= traversed
            requests = [
                reflection_pb2.ServerReflectionRequest(file_by_filename=dep)
                for dep in deps
            ]
            # prevent unsatisfied deps from looping forever
            traversed.update(deps)

        names = deque(protos.keys())
        traversed = set()
        while names:
            name = names[0]
            traversed.add(name)
            # raises KeyError if unsatisfied dep:
            proto = protos[name]
            deps = set(proto.dependency) - traversed
            if deps:
                names = deque(x for x in names if x not in deps)
                names.extendleft(deps)
            else:
                del names[0]
                self.pool.Add(proto)

        return protos.keys()

    def file_descriptor(self, name):
        """Get file descriptor for a proto file that has been loaded.

        Args:
            name (str): The file name of the .proto file, including path.

        Returns:
             google.protobuf.descriptor.FileDescriptor: The descriptor for the
                specified file.

        Raises:
            KeyError: File has not been loaded.
        """
        return self.pool.FindFileByName(name)

    def message_from_descr(self, proto):
        """Get message class from message descriptor.

        Args:
            proto (google.protobuf.descriptor.Descriptor): Message descriptor.
        """
        return message_factory.GetMessageClass(proto)

    def gen_stub_class(self, service, method_protos):
        """Get service stub class from service descriptor.

        Args:
            service (google.protobuf.descriptor.ServiceDescriptor): Service
                descriptor.
            method_protos (iterable): The
                `google.protobuf.descriptor_pb2.MethodDescriptorProto` objects
                for the methods in the service.
        """
        stub_methods = []
        dep_descrs = []
        for method_proto in method_protos:
            method = service.methods_by_name[method_proto.name]
            channel_call = "_".join(
                "stream" if x else "unary" for x in
                [method_proto.client_streaming, method_proto.server_streaming])
            dep_descrs.extend([method.input_type, method.output_type])
            stub_methods.append(
                (method.name, channel_call,
                 "/{}/{}".format(service.full_name, method.name),
                 self.message_from_descr(method.input_type).SerializeToString,
                 self.message_from_descr(method.output_type).FromString))
        class_name = service.name + "Stub"
        return type(class_name, (), {
            "_methods": stub_methods,
            "__init__": __stub_init__
        }), dep_descrs


class GrpcReflectionClient():
    """Client to discover protocol types via gRPC service reflection.

    Protocol files must be loaded via `load_protocols` prior to getting any of
    the class types.

    Note that for nested types (fields, oneofs, etc) other than nested message
    types, the top-level message type should be requested and the nested types
    can be accessed on it via attribute.
    """

    def __init__(self):
        self._engine = GrpcReflectionEngine()

    def load_protocols(self, channel, filenames=None, symbols=None):
        """Load a set of proto files to use later for protocol types.

        Load one or more specified files and/or files containing specified
        symbols, as well as the transitive dependency files of those files.

        If neither `filenames` nor `symbols` is specified, then files
        containing the symbols for all services advertised by the RPC server,
        other than the reflection service, are loaded.

        This method may be called multiple times to load additional proto
        files; however, it is not recommended to use channels that point to
        different servers unless it is known that the symbols either do not
        overlap or contain the same exact version across all servers.

        The caller-supplied channel is only used within the context of this
        method call, so the caller is free to close it afterwards.

        Args:
            channel (grpc.Channel): The RPC channel to use.
            filenames (iterable[str]): Optional. Proto file names, with path,
                to request.
            symbols (iterable[str]): Optional. Fully qualified symbol names
                for which to request proto files.

        Raises:
            ServiceError: Reflection service responded with an error.
            grpc.RpcError: Lower level RPC error.
        """
        return self._engine.load_protocols(channel, filenames, symbols)

    def service_stub_class(self, name):
        """Get the stub class for the specified service type.

        Args:
            name: Fully qualified name of the protocol service.

        Raises:
            KeyError: File defining service has not been loaded.
        """
        service = self._engine.pool.FindServiceByName(name)
        methods = self._engine.methods_by_file[service.file.name][service.name]
        return self._engine.gen_stub_class(service, methods)[0]

    def message_class(self, name):
        """Get the class for the specified message type.

        Args:
            name: Fully qualified name of the protocol message.

        Raises:
            KeyError: File defining message has not been loaded.
        """
        return self._engine.message_from_descr(
            self._engine.pool.FindMessageTypeByName(name))

    def enum_class(self, name):
        """Get the class for the specified enum type.

        Args:
            name: Fully qualified name of the protocol enum.

        Raises:
            KeyError: File defining enum has not been loaded.
        """
        return enum_from_descr(self._engine.pool.FindEnumTypeByName(name))



#!/usr/bin/python3
"""Server for the ridiculous Subtraction service.

This is meant for running this project's examples against.

Note that `arithmetic.subtract_pb2` and `arithmetic.subtract_pb2_grpc` are
deliberately kept out of the import path for the reflection-based example
clients, but the service needs them.
"""
from concurrent import futures
import logging

import grpc
from grpc_reflection.v1alpha import reflection

from arithmetic import subtract_pb2
from arithmetic import subtract_pb2_grpc


class Subtraction(subtract_pb2_grpc.SubtractionServicer):

    def SubtractOne(self, request, context):
        return subtract_pb2.Difference(number=request.number - 1)


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    subtract_pb2_grpc.add_SubtractionServicer_to_server(Subtraction(), server)
    service_names = (
        subtract_pb2.DESCRIPTOR.services_by_name['Subtraction'].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(service_names, server)
    server.add_insecure_port('[::]:19002')
    server.start()
    server.wait_for_termination()


if __name__ == '__main__':
    logging.basicConfig()
    serve()



import importlib

import pytest
from yagrc import importer as yagrc_importer

# See comments in Add_One.proto about terrible naming


@pytest.fixture
def grpc_importer(grpc_channel):
    importer = yagrc_importer.GrpcImporter()
    importer.configure(grpc_channel, filenames=["Testing_protos/Add_One.proto"])
    yield importer
    importer.deconfigure()


def test_AddOne(grpc_channel, grpc_importer):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.AddOne(AddTypes_pb2.Addend(number=5))
    assert response.number == 6


def test_AddOnes(grpc_channel, grpc_importer):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    responses = stub.AddOnes(AddTypes_pb2.Addend(number=5))
    for i, response in enumerate(responses):
        assert response.number == 6 + i
    assert i == 4


def test_AddsOne(grpc_channel, grpc_importer):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.AddsOne((AddTypes_pb2.Addend(number=x) for x in range(5)))
    assert response.number == 0 + 1 + 2 + 3 + 4 + 1


def test_AddsOnes(grpc_channel, grpc_importer):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    responses = stub.AddsOnes((AddTypes_pb2.Addend(number=x) for x in range(5)))
    for i, response in enumerate(responses):
        assert response.number == i + 1
    assert i == 4


def test_CountQuery(grpc_channel, grpc_importer):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.Status(AddTypes_pb2.StatusQuery(count={}))
    assert response.count.number > 0


def test_StatusQuery(grpc_channel, grpc_importer):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.Status(AddTypes_pb2.StatusQuery(state={}))
    assert response.state.states[
        "sum_engine"].state == AddTypes_pb2.ServiceState.READY
    assert AddTypes_pb2.ServiceState.Name(
        response.state.states["difference_engine"].state
    ) == "BUSY_COMPUTING_ONE"
    assert response.state.states["chaos_engine"].state == AddTypes_pb2._UNKNOWN


def test_DESCRIPTOR(grpc_importer):
    from Testing_protos import Add_One_pb2
    assert Add_One_pb2.DESCRIPTOR.name == "Testing_protos/Add_One.proto"


def test_invalidate_caches(grpc_importer):
    from Testing_protos import Add_One_pb2
    assert grpc_importer._finder._imported
    importlib.invalidate_caches()
    assert not grpc_importer._finder._imported


def test_path_prepend(grpc_channel, grpc_importer):
    grpc_importer.configure(grpc_channel, path_prepend=True)
    from Testing_protos import Add_One_pb2


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_deconfigure(grpc_channel):
    importer = yagrc_importer.GrpcImporter()
    importer.configure(grpc_channel, filenames=["Testing_protos/Add_One.proto"])
    from Testing_protos import Add_One_pb2
    importer.deconfigure()
    from Testing_protos import Add_One_pb2


# This is just to make sure following test is meaningful
@pytest.mark.xfail(raises=NameError, strict=True)
def test_no_import(grpc_importer):
    assert Testing_protos.AddAltTypes_pb2.Addend.DESCRIPTOR.name == "Addend"


# While it is probably bad practice to rely on this, the generated _pb2 modules
# import all their transitive dependencies into the package namespace, and the
# yagrc importer deliberately mimics that behavior.
def test_indirect_import(grpc_importer):
    import Testing_protos.Add_One_pb2_grpc
    assert Testing_protos.AddAltTypes_pb2.Addend.DESCRIPTOR.name == "Addend"


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_bad_import(grpc_channel, grpc_importer):
    from Testing_protos import bad_import_pb2
    assert False



"""An import provider that mimics generated grpc protocol module files.

This modules builds on top of `yagrc.reflector` to provide a drop-in
replacement for the generated _pb2 and _pb2_grpc modules supporting the
gRPC services advertised by a server that supports service reflection.
"""
import importlib
import importlib.abc
import importlib.machinery
import sys

from yagrc import reflector as yagrc_reflector


def _proto_basename(name):
    return (name[:-6] if name.endswith(".proto") else name).replace("/", ".")


class _DummyLoader(importlib.abc.Loader):  # pylint: disable=abstract-method

    def __init__(self, imported_set):
        super().__init__()
        self._imported = imported_set

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        if self._imported is not None:
            self._imported.add(module.__name__)


class _DynamicPb2Loader(importlib.abc.Loader):  # pylint: disable=abstract-method

    def __init__(self, imported_set, filename, reflector):
        super().__init__()
        self._imported = imported_set
        self._filename = filename
        self._reflector = reflector

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        self._imported.add(module.__name__)
        _exec_pb2_module(module, self._reflector, self._filename)


def _exec_pb2_module(module, reflector, filename):
    file_descr = reflector.file_descriptor(filename)
    module.DESCRIPTOR = file_descr
    # mimic _pb2 module dependency import behavior
    for dep_descr in file_descr.dependencies:
        dep_mod = importlib.import_module(
            _proto_basename(dep_descr.name) + "_pb2")
        if dep_descr in file_descr.public_dependencies:
            module.__dict__.update((key, val)
                                   for key, val in dep_mod.__dict__.items()
                                   if not key.startswith("_"))
    for name, message in file_descr.message_types_by_name.items():
        setattr(module, name, reflector.message_from_descr(message))
    for name, enum in file_descr.enum_types_by_name.items():
        setattr(module, name, yagrc_reflector.enum_from_descr(enum))
        for value in enum.values:
            # seems crazy to do this blindly, but the generated files do so
            setattr(module, value.name, value.number)


class _DynamicPb2GrpcLoader(importlib.abc.Loader):  # pylint: disable=abstract-method

    def __init__(self, imported_set, filename, reflector):
        super().__init__()
        self._imported = imported_set
        self._filename = filename
        self._reflector = reflector

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        self._imported.add(module.__name__)
        _exec_pb2_grpc_module(module, self._reflector, self._filename)


def _exec_pb2_grpc_module(module, reflector, filename):
    file_descr = reflector.file_descriptor(filename)
    dep_filenames = set()
    method_proto_map = reflector.methods_by_file[filename]
    for service_name, method_protos in method_proto_map.items():
        service = file_descr.services_by_name[service_name]
        stub_class, dep_descrs = reflector.gen_stub_class(
            service, method_protos)
        stub_class.__module__ = module.__name__
        setattr(module, stub_class.__name__, stub_class)
        dep_filenames.update(descr.file.name for descr in dep_descrs)
    # mimic _pb2_grpc module dependency behavior
    for dep_filename in dep_filenames:
        importlib.import_module(_proto_basename(dep_filename) + "_pb2")


class _DynamicGrpcFinder(importlib.abc.MetaPathFinder):

    def __init__(self, reflector):
        super().__init__()
        self._package_modules = set()
        self._pb2_modules = {}
        self._pb2_grpc_modules = {}
        self._imported = set()
        self._reflector = reflector

    def configure_files(self, filenames):
        """Add modules and their containing package paths."""
        for filename in filenames:
            base = _proto_basename(filename)
            self._pb2_modules[base + "_pb2"] = filename
            self._pb2_grpc_modules[base + "_pb2_grpc"] = filename
            while base:
                base, _, _ = base.rpartition(".")
                if base:
                    self._package_modules.add(base)

    def reset_files(self):
        """Remove previously added modules and package paths."""
        self.invalidate_caches()
        self._package_modules.clear()
        self._pb2_modules.clear()
        self._pb2_grpc_modules.clear()

    def find_spec(self, fullname, path, target=None):  # pylint: disable=unused-argument
        if fullname in self._package_modules:
            return importlib.machinery.ModuleSpec(fullname,
                                                  _DummyLoader(self._imported),
                                                  is_package=True)
        if fullname in self._pb2_modules:
            return importlib.machinery.ModuleSpec(
                fullname,
                _DynamicPb2Loader(self._imported, self._pb2_modules[fullname],
                                  self._reflector),
                is_package=False)
        if fullname in self._pb2_grpc_modules:
            return importlib.machinery.ModuleSpec(
                fullname,
                _DynamicPb2GrpcLoader(self._imported,
                                      self._pb2_grpc_modules[fullname],
                                      self._reflector),
                is_package=False)
        return None

    def invalidate_caches(self):
        # can't do much more without requiring configure to be called again
        for name in self._imported:
            if name in sys.modules:
                del sys.modules[name]
        self._imported.clear()


class GrpcImporter():
    """Public API for supporting dynamic import of protocol modules.

    Note that direct usage of this class is not required when using
    `add_lazy_packages` and `resolve_lazy_imports`.
    """

    def __init__(self):
        self.reflector = yagrc_reflector.GrpcReflectionEngine()
        self._finder = _DynamicGrpcFinder(self.reflector)

    def configure(self, channel, filenames=None, path_prepend=False):
        """Load a set of proto files and enable their dynamic module imports.

        Load one or more specified files, as well as the transitive dependency
        files of those files. If `filenames` is not specified, then files
        containing the symbols for all services advertised by the RPC server,
        other than the reflection service, are loaded.

        For each proto file loaded, corresponding _pb2 and _pb2_grpc modules
        will be made available in the import path. The module names and
        package path will be the same as that generated by the protoc tool
        for the proto files used on the server, which is a function of
        directory layout of the proto files.

        This method may be called multiple times to load additional proto
        files.

        The caller-supplied channel is only used within the context of this
        method call, so the caller is free to close it afterwards.

        Args:
            channel (grpc.Channel): The RPC channel to use.
            filenames (iterable[str]): Optional. Proto file names, with path,
                to request.
            path_prepend (bool): Optional. If set to True, places the dynamic
                proto modules ahead of any other import path, otherwise places
                them after other import paths.

        Raises:
            yagrc.reflector.ServiceError: Reflection service responded with an
                error.
            grpc.RpcError: Lower level RPC error.
        """
        loaded = self.reflector.load_protocols(channel, filenames=filenames)
        self._finder.configure_files(loaded)

        if self._finder in sys.meta_path:
            sys.meta_path.remove(self._finder)
        if path_prepend:
            sys.meta_path.insert(0, self._finder)
        else:
            sys.meta_path.append(self._finder)

    def deconfigure(self):
        """Remove support for subsequent dynamic module imports.

        Dynamic modules that have already been imported remain usable by the
        modules that imported them, but new imports will fail unless the
        modules are provided elsewhere in the module path, or
        `GrpcImporter.configure` is called again.

        This is not normally required, but without it, the entry in the import
        meta path will remain in place, even if the `GrpcImporter` object that
        installed it is destroyed.
        """
        if self._finder in sys.meta_path:
            sys.meta_path.remove(self._finder)
        self._finder.reset_files()


def __lazy_getattr__(name):
    raise AttributeError("attribute access on unresolved lazy module: " + name)


class _LazyLoader(importlib.abc.Loader):  # pylint: disable=abstract-method

    def __init__(self, imported_list):
        super().__init__()
        self._imported = imported_list

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        self._imported.append(module)
        module.__getattr__ = __lazy_getattr__


class _LazyGrpcFinder(importlib.abc.MetaPathFinder):

    def __init__(self):
        super().__init__()
        self._package_modules = set()
        self.pb2_imports = []
        self.pb2_grpc_imports = []

    def add_package(self, package):
        """Add a package path to list of auto-imports"""
        while package:
            self._package_modules.add(package)
            package, _, _ = package.rpartition(".")

    def find_spec(self, fullname, path, target=None):  # pylint: disable=unused-argument
        if fullname in self._package_modules:
            return importlib.machinery.ModuleSpec(fullname,
                                                  _DummyLoader(None),
                                                  is_package=True)

        package, _, module = fullname.rpartition(".")
        if package in self._package_modules:
            if module.endswith("_pb2"):
                return importlib.machinery.ModuleSpec(fullname,
                                                      _LazyLoader(
                                                          self.pb2_imports),
                                                      is_package=False)
            if module.endswith("_pb2_grpc"):
                return importlib.machinery.ModuleSpec(
                    fullname,
                    _LazyLoader(self.pb2_grpc_imports),
                    is_package=False)

        return None


# for simplicity of usage, lazy import uses singletons
_lazy_finder = _LazyGrpcFinder()
_lazy_importer = GrpcImporter()


def add_lazy_packages(packages):
    """Configure one or more package paths for lazy resolution.

    For each package name, allows any module within that package and with name
    ending in "_pb2" or "_pb2_grpc" to be imported regardless of whether its
    corresponding proto file has been loaded via `GrpcImporter.configure` or
    is otherwise present in the module import path.

    `resolve_lazy_imports` must be called later to make the classes and other
    attributes in the module available for use.

    Args:
        packages (iterable[str]): Fully qualified package names.
    """
    for package in packages:
        _lazy_finder.add_package(package)
    if _lazy_finder not in sys.meta_path:
        sys.meta_path.append(_lazy_finder)


def resolve_lazy_imports(channel):
    """Make lazily loaded modules available for use.

    Args:
        channel (grpc.Channel): The RPC channel to use for reflection service.

    Raises:
        yagrc.reflector.ServiceError: Reflection service responded with an
            error.
        grpc.RpcError: Lower level RPC error.
    """
    if not _lazy_finder.pb2_imports and not _lazy_finder.pb2_grpc_imports:
        return

    pb2_modules = tuple(
        (module.__name__[:-4].replace(".", "/") + ".proto", module)
        for module in _lazy_finder.pb2_imports)

    pb2_grpc_modules = tuple(
        (module.__name__[:-9].replace(".", "/") + ".proto", module)
        for module in _lazy_finder.pb2_grpc_imports)

    orig_meta_path = sys.meta_path.copy()
    if _lazy_finder in sys.meta_path:
        sys.meta_path.remove(_lazy_finder)

    try:
        filenames = set(x[0] for x in pb2_modules + pb2_grpc_modules)
        _lazy_importer.configure(channel, filenames=filenames)
        for filename, module in pb2_modules:
            del module.__getattr__
            _exec_pb2_module(module, _lazy_importer.reflector, filename)
        for filename, module in pb2_grpc_modules:
            del module.__getattr__
            _exec_pb2_grpc_module(module, _lazy_importer.reflector, filename)
    except:
        sys.meta_path = orig_meta_path
        raise

    _lazy_finder.pb2_imports.clear()
    _lazy_finder.pb2_grpc_imports.clear()
    sys.meta_path.append(_lazy_finder)



import pytest

# Verify failure when nothing is done to make Testing_protos package available
# in module path.

# See comments in Add_One.proto about terrible naming

# Uncomment the following lines to verify that the test cases would otherwise
# pass if not for the import errors. (this will cause the test suite to fail)
#import os
#import sys
#sys.path.insert(
#    0, os.path.join(os.path.dirname(os.path.abspath(__file__)), "service"))


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_AddOne(grpc_channel):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.AddOne(AddTypes_pb2.Addend(number=5))
    assert response.number == 6


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_AddOnes(grpc_channel):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    responses = stub.AddOnes(AddTypes_pb2.Addend(number=5))
    for i, response in enumerate(responses):
        assert response.number == 6 + i
    assert i == 4


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_AddsOne(grpc_channel):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.AddsOne((AddTypes_pb2.Addend(number=x) for x in range(5)))
    assert response.number == 0 + 1 + 2 + 3 + 4 + 1


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_AddsOnes(grpc_channel):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    responses = stub.AddsOnes((AddTypes_pb2.Addend(number=x) for x in range(5)))
    for i, response in enumerate(responses):
        assert response.number == i + 1
    assert i == 4


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_CountQuery(grpc_channel):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.Status(AddTypes_pb2.StatusQuery(count={}))
    assert response.count.number > 0


@pytest.mark.xfail(raises=ImportError, strict=True)
def test_StatusQuery(grpc_channel):
    from Testing_protos import Add_One_pb2_grpc
    from Testing_protos import AddTypes_pb2
    stub = Add_One_pb2_grpc.AdditionStub(grpc_channel)
    response = stub.Status(AddTypes_pb2.StatusQuery(state={}))
    assert response.state.states[
        "sum_engine"].state == AddTypes_pb2.ServiceState.READY
    assert AddTypes_pb2.ServiceState.Name(
        response.state.states["difference_engine"].state
    ) == "BUSY_COMPUTING_ONE"
    assert response.state.states["chaos_engine"].state == AddTypes_pb2._UNKNOWN
